#ifndef	_CoincidencedHit_h_
#define	_CoincidencedHit_h_

#include <vector>
#include <algorithm>

#include <assert.h>
#include <TCollection.h>
#include <TObjArray.h>
#include "Hit.h"


/*** Internally used in FillList2 ******************/
class timeinfo {
public:
    int time;
    int op;	/* +1: outer hit opens gate */
		/* -1: outer hit closes gate */
		/*  0: gate state is not changed (used by innter hit) */
};

class timeinfo_cmp_functor {
public:
    bool operator()(const timeinfo & a, const timeinfo & b) const {
	if (b.time > a.time)
	    return true;

	if (b.time == a.time) {
	    if (a.op > b.op)
		return true;
	}

	return false;
    };
};
/***************************************************/

class CoincidencedHit : public TObject {
private:
    Hit fHit1;
    Hit fHit2;
public:
    CoincidencedHit() : fHit1(), fHit2() { };
    CoincidencedHit(Hit * h1, Hit * h2) : fHit1(*h1), fHit2(*h2) { };
    CoincidencedHit(Hit h1, Hit h2) : fHit1(h1), fHit2(h2) { };
    virtual ~CoincidencedHit() { };

    virtual void Print(Option_t * opt = "") const {
	printf("%s: Ch=%d,%d Time=%d,%d Gap=%d\n",
	    ClassName(),
	    fHit1.GetCh(), fHit2.GetCh(),
	    fHit1.GetTime(), fHit2.GetTime(),
	    abs(fHit2.GetTime() - fHit1.GetTime()));
    };

    const Hit & GetHit1() const { return fHit1; };
    const Hit & GetHit2() const { return fHit2; };

    Int_t GetGap() {
	return abs(fHit1.GetTime() - fHit2.GetTime());
    };

    static void FillList(TCollection & src1, TCollection & src2, Int_t gap, TCollection & result) {
	TObjArray joined;

	result.SetOwner(kTRUE);

	joined.AddAll(&src1);
	joined.AddAll(&src2);

	joined.Sort();
	joined.Compress(); /* remove empty slot */

	// joined.Print();

	Int_t n = joined.GetEntries() - 1;

	for (int i=0; i<n; i++) {
	    TObject * o1 = joined.At(i);
	    TObject * o2 = joined.At(i+1);

	    assert(o1);
	    assert(o2);
	    assert(o1->InheritsFrom(Hit::Class()));
	    assert(o2->InheritsFrom(Hit::Class()));

	    Hit * hit1 = (Hit *)o1;
	    Hit * hit2 = (Hit *)o2;

	    if (hit1->GetCh() == hit2->GetCh()) {
		joined.Remove(hit1);
		continue;
	    }

	    if ((hit2->GetTime() - hit1->GetTime()) > gap) {
		joined.Remove(hit1);
		continue;
	    }

	    result.Add(new CoincidencedHit(hit1, hit2));
	}
    };

    static void FillList2(TCollection & inner_list, TCollection & outer_list,
		Int_t gap, TCollection & result) {
	const Int_t w = gap / 2;
	const Int_t n_inner(inner_list.GetEntries());
	const Int_t n_outer(outer_list.GetEntries());

	if (n_inner == 0)
	    return;

	if (n_outer == 0)
	    return;

	Int_t vector_size = n_inner + n_outer * 2;

	std::vector<timeinfo> t(vector_size);

	int i = 0;

	int inner_ch = -1, outer_ch = -1;

	TIter inner_iter(&inner_list);

	while (TObject * o = inner_iter()) {
	    assert(o->InheritsFrom(Hit::Class()));
	    Hit * h = (Hit *)o;

	    if (inner_ch == -1)
		inner_ch = h->GetCh();

	    t[i].time = h->GetTime();
	    t[i].op = 0;

	    i++;
	}


	TIter outer_iter(&outer_list);

	while (TObject * o = outer_iter()) {
	    assert(o->InheritsFrom(Hit::Class()));
	    Hit * h = (Hit *)o;

	    if (outer_ch == -1)
		outer_ch = h->GetCh();

	    t[i].time = h->GetTime() - w;
	    t[i].op = 1;	/* Gate Open */
	    i++;

	    t[i].time = h->GetTime() + w;
	    t[i].op = -1;	/* Gate Close */
	    i++;
	}

	assert(i == vector_size);

	std::sort(t.begin(), t.end(), timeinfo_cmp_functor());

	result.SetOwner(kTRUE);

	int state = 0;

	for (int i=0; i<vector_size; i++) {
	    Int_t last_outer_time;

	    switch (t[i].op) {
	    case +1: /* Gate Open */
		last_outer_time = t[i].time + w;
	    case -1: /* Gate Close */
		state += t[i].op;
		break;
	    case 0: /* Inner hit */
		if (state > 0) { /* If Gate is opened */
		    Hit inner(inner_ch, t[i].time);
		    Hit outer(outer_ch, last_outer_time);
		    result.Add(new CoincidencedHit(inner, outer));
		}
		break;
	    }

	}
    };


    static TCollection * GenerateList(TCollection * src1, TCollection * src2, Int_t gap) {
	TObjArray * result = new TObjArray;
	assert(src1);
	assert(src2);

	FillList(*src1, *src2, gap, *result);

	return result;
    };


    ClassDef(CoincidencedHit, 1)
};

#endif
